import json
import logging
from typing import Optional

import requests

logger = logging.getLogger(__name__)


class SodaCloudClient:
    def __init__(
        self,
        api_key_id: str,
        api_key_secret: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
        token: Optional[str] = None,
    ):
        self.host = host
        self.port = f":{port}" if port else ""
        self.api_url = f"https://{self.host}{self.port}/api"
        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.token: Optional[str] = token

    def insert_scan_results(self, scan_results):
        scan_results["type"] = "sodaCoreInsertScanResults"
        return self._execute_command(scan_results)

    def _execute_command(self, command: dict):
        logger.debug(f"executing {command}")
        return self._execute_request("command", command, False)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool):
        logger.debug(f"> /api/{request_type} {json.dumps(request_body, indent=2)}")
        request_body["token"] = self._get_token()
        response = requests.post(f"{self.api_url}/{request_type}", json=request_body)
        response_json = response.json()
        if response.status_code == 401 and not is_retry:
            logger.debug(f"Authentication failed. Probably token expired. Re-authenticating...")
            self.token = None
            response_json = self._execute_request(request_type, request_body, True)
        else:
            assert (
                response.status_code == 200
            ), f"Request failed with status {response.status_code}: {json.dumps(response_json, indent=2)}"
        return response_json

    def _get_token(self):
        if not self.token:
            login_command = {"type": "login"}
            if self.api_key_id and self.api_key_secret:
                logger.debug("> /api/command (login with API key credentials)")
                login_command["apiKeyId"] = self.api_key_id
                login_command["apiKeySecret"] = self.api_key_secret
            else:
                raise RuntimeError("No API KEY and/or SECRET provided ")

            login_response = requests.post(f"{self.api_url}/command", json=login_command)

            if login_response.status_code != 200:
                raise AssertionError(f"< {login_response.status_code} Login failed: {login_response.content}")
            login_response_json = login_response.json()
            self.token = login_response_json.get("token")
            assert self.token, "No token in login response?!"
            logger.debug("< 200 (login ok, token received)")
        return self.token



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional


@dataclass
class Feedback:
    is_anomaly: bool
    is_correctly_classified: bool


@dataclass
class TimedValue:
    timestamp: datetime
    value: float
    feedback: Optional[Feedback]


@dataclass
class AnomalyInput:
    timed_values: List[TimedValue]


@dataclass
class AnomalyOutput:
    is_anomaly: bool
    anomaly_score: float


class AnomalyDetector(ABC):
    @abstractmethod
    def evaluate(self, anomaly_input: AnomalyInput) -> AnomalyOutput:
        pass



from typing import Optional


class HistoricDescriptor:
    def __init__(
        self,
        metric: "Metric" = None,
        change_over_time_cfg: "ChangeOverTimeCfg" = None,
        anomaly_values: Optional[int] = None,
    ):
        from soda.sodacl.change_over_time_cfg import ChangeOverTimeCfg

        self.metric = metric
        self.change_over_time_cfg: ChangeOverTimeCfg = change_over_time_cfg
        self.anomaly_values: Optional[int] = anomaly_values

    def __hash__(self):
        return hash((self.metric, self.change_over_time_cfg))

    def __eq__(self, other):
        if type(other) != HistoricDescriptor:
            return False
        if self is other:
            return True
        return (
            self.metric == other.metric
            and self.change_over_time_cfg == other.change_over_time_cfg
            and self.anomaly_values == self.anomaly_values
        )

    def to_jsonnable(self):
        jsonnable = {}
        if self.metric:
            jsonnable["metric_id"] = self.metric.identity
        if self.change_over_time_cfg:
            jsonnable["change_over_time"] = self.change_over_time_cfg.to_jsonnable()
        if self.anomaly_values is not None:
            jsonnable["anomaly_values"] = self.anomaly_values
        return jsonnable



from __future__ import annotations

from soda.execution.metric import Metric
from soda.execution.metric_check import MetricCheck
from soda.soda_cloud.historic_descriptor import HistoricDescriptor
from soda.sodacl.metric_check_cfg import MetricCheckCfg

KEY_HISTORIC_ANOMALY_VALUES = "historic_anomaly_values"


class AnomalyMetricCheck(MetricCheck):
    def __init__(
        self,
        check_cfg: MetricCheckCfg,
        data_source_scan: DataSourceScan,
        partition: Partition | None = None,
        column: Column | None = None,
    ):
        super().__init__(
            check_cfg=check_cfg,
            data_source_scan=data_source_scan,
            partition=partition,
            column=column,
        )

        metric_check_cfg: MetricCheckCfg = self.check_cfg
        metric_name = metric_check_cfg.metric_name
        metric = self.metrics[metric_name]

        self.historic_descriptors[KEY_HISTORIC_ANOMALY_VALUES] = HistoricDescriptor(
            metric=metric,
            anomaly_values=90,
        )

    def evaluate(self, metrics: dict[str, Metric], historic_values: dict[str, object]):
        metric_value = self.get_metric_value()
        historic_anomaly_values = historic_values.get(KEY_HISTORIC_ANOMALY_VALUES)

        if historic_anomaly_values:
            self.check_value = self.compute_anomaly_score(historic_anomaly_values, metric_value)
            self.set_outcome_based_on_check_value()

            # put all diagnostics into a member field like this:
            self.anomaly_values = {}

        else:
            self.logs.info("Skipping metric check eval because there is not enough historic data yet")

    def compute_anomaly_score(self, historic_anomaly_values: list[dict], metric_value: int | float) -> float:
        data_timestamp = self.data_source_scan.scan._data_timestamp

        # TODO invoke the anomaly detection algorithm dynamically as it is in an extension module
        # ensure appropriate error log if the extension module is not installed
        # The historic_anomaly_values look like this:
        # [
        #     {
        #         'identity': 'metric-test_change_over_time.py::test_anomaly_detection-postgres-SODATEST_Customers_b7580920-row_count',
        #         'data_time': datetime.datetime(2022, 3, 8, 19, 40, 5, 880298),
        #         'value': 10
        #     },
        #     {
        #         'identity': 'metric-test_change_over_time.py::test_anomaly_detection-postgres-SODATEST_Customers_b7580920-row_count',
        #         'data_time': datetime.datetime(2022, 3, 7, 19, 40, 5, 880298),
        #         'value': 10
        #     }, ...
        # ]

        anomaly_score = 0.8

        return anomaly_score

    def get_cloud_diagnostics_dict(self) -> dict:
        cloud_diagnostics = super().get_cloud_diagnostics_dict()
        if self.historic_diff_values:
            cloud_diagnostics["anomaly_values"] = self.anomaly_values

    def get_log_diagnostic_dict(self) -> dict:
        log_diagnostics = super().get_log_diagnostic_dict()
        if self.historic_diff_values:
            log_diagnostics.update(self.anomaly_values)
        return log_diagnostics
